/*
**  Copyright (C) 1993-1996, OFFIS, Oldenburg University and CERIUM
**  
**  This software and supporting documentation were
**  developed by 
**  
**    Institut OFFIS
**    Bereich Kommunikationssysteme
**    Westerstr. 10-12
**    26121 Oldenburg, Germany
**    
**    Fachbereich Informatik
**    Abteilung Prozessinformatik
**    Carl von Ossietzky Universitaet Oldenburg 
**    Ammerlaender Heerstr. 114-118
**    26111 Oldenburg, Germany
**    
**    CERIUM
**    Laboratoire SIM
**    Faculte de Medecine
**    2 Avenue du Pr. Leon Bernard
**    35043 Rennes Cedex, France
**  
**  for CEN/TC251/WG4 as a contribution to the Radiological 
**  Society of North America (RSNA) 1993 Digital Imaging and 
**  Communications in Medicine (DICOM) Demonstration.
**  
**  THIS SOFTWARE IS MADE AVAILABLE, AS IS, AND NEITHER OFFIS,
**  OLDENBURG UNIVERSITY NOR CERIUM MAKE ANY WARRANTY REGARDING 
**  THE SOFTWARE, ITS PERFORMANCE, ITS MERCHANTABILITY OR 
**  FITNESS FOR ANY PARTICULAR USE, FREEDOM FROM ANY COMPUTER 
**  DISEASES OR ITS CONFORMITY TO ANY SPECIFICATION.  THE 
**  ENTIRE RISK AS TO QUALITY AND PERFORMANCE OF THE SOFTWARE   
**  IS WITH THE USER. 
**  
**  Copyright of the software and supporting documentation
**  is, unless otherwise stated, jointly owned by OFFIS,
**  Oldenburg University and CERIUM and free access is hereby
**  granted as a license to use this software, copy this
**  software and prepare derivative works based upon this
**  software. However, any distribution of this software
**  source code or supporting documentation or derivative
**  works (source code and supporting documentation) must
**  include the three paragraphs of this copyright notice. 
** 
*/
/*	
**			ctndisp.c
**
**	CTNDISP is a program for displaying images of arbitrary 
**	size that have been enqueued (using the snq routines) to
**	the Study Name Queue (SNQ).  CTNDISP's principal purpose 
**	is to display images as evidence of a successful
**	transfer of an image by an SCP process.  Images can also
**	be queued via a simple TTY interface (see SNQ) for demonstration
**	and diagnostic purposes.  
**	
**	ACR image names are retrieved from a shared memory segment that
**	is created by the SCP process.  This shared memory segment,
**	known as the Study Name Queue (SNQ), is a FIFO queue
**	on which the IDX image database names are enqueued.
**
**	CTNDISP always displays the first image of any given study.
**	Images are displayed using gray-scale lookup table.  The patient
**	name, date, modality, and study are displayed above each image.  The
**	node name is displayed at the top.  
**
**	 
** $Log: ctndisp.cc,v $
** Revision 1.12  2000/03/03 14:17:54  meichel
** Implemented library support for redirecting error messages into memory
**   instead of printing them to stdout/stderr for GUI applications.
**
** Revision 1.11  2000/02/23 15:11:18  meichel
** Corrected macro for Borland C++ Builder 4 workaround.
**
** Revision 1.10  2000/02/01 17:16:08  meichel
** Adapted ctndisp to OFCommandLine class.
**
** Revision 1.9  2000/02/01 10:49:06  meichel
** Avoiding to include <stdlib.h> as extern "C" on Borland C++ Builder 4,
**   workaround for bug in compiler header files.
**
** Revision 1.8  1999/05/04 12:36:43  meichel
** Including ofstring.h to avoid prototype conflicts on HP-UX 9 with gcc.
**
** Revision 1.7  1997/10/22 09:21:06  meichel
** Corrected check for presence of shared memory system calls in ctndisp.
**   Needed for AIX 3.2 (IBM xlC 1.0) where shmctl and shmdt are present, but shmat is not.
**
** Revision 1.6  1997/07/24 13:10:42  andreas
** - Removed Warnings from SUN CC 2.0.1
**
** Revision 1.5  1997/05/30 12:29:30  andreas
** - include version number in every usage String
**
** Revision 1.4  1997/05/29 15:52:47  meichel
** Added constant for dcmtk release date in dcuid.h.
** All dcmtk applications now contain a version string
** which is displayed with the command line options ("usage" message)
** and which can be queried in the binary with the "ident" command.
**
** Revision 1.3  1997/05/22 13:23:41  hewett
** Modified the test for presence of a data dictionary to use the
** method DcmDataDictionary::isDictionaryLoaded().
**
** Revision 1.2  1996/04/18 15:29:40  meichel
** Version for CAR'96
**
** Revision 1.1  1996/03/28 17:06:36  meichel
** first commit with new file structure
**
** Revision 2.2  1995/05/18 11:57:01  hewett
** Ran code through a formatter.  I now a little more readable.
**
 * Revision 2.1  1995/02/02  12:49:25  hewett
 * Corrected spelling of Ossietzky
 *
 * Revision 2.0  1994/11/17  09:44:08  meichel
 * Version 2.0 for RSNA 1994 demonstration.
 * New features: Metaheader support, new SOP classes, lots of bugs fixed.
 *
 * Revision 1.15  1994/03/17  09:27:46  hewett
 * Updated copyright notice.
 *
 * Revision 1.14  1994/03/16  16:17:27  hewett
 * Reduced the number of gray levels.  Get really wierd results on
 * Solaris2 otherwise.
 *
 * Revision 1.13  1993/11/22  12:31:49  hewett
 * Removed 10mb queue input buffer.  No longer necessary since snq fixed.
 *
 * Revision 1.12  1993/11/19  15:11:33  jens
 * Fixed some bugs.
 *
 * Revision 1.11  1993/11/18  20:17:01  jens
 * Multi processes Version of ctndisp.
 * If you don`t want forked image conversion, then use -noMulti switch
 *
 * Revision 1.10  1993/11/16  12:24:30  jens
 * Added switch -2images to ctndisp. Only 2 Imageareas are used.
 * Modified the Assiciation window refresh. (less flicker)
 * Added a image count information.
 * Fasten XputImage.
 * Added code for the killassoc command.
 *
 * Revision 1.9  1993/11/10  17:42:55  jens
 * Added a new flag: -half
 * with ctndisp -half the window uses a square half of the screen
 *
 * Revision 1.8  1993/11/01  21:02:16  jens
 * Better error resistent Version in Tag reading section.
 *
 * Revision 1.7  1993/10/14  10:46:08  hewett
 * Added Euro Header
 *
 * Revision 1.6  1993/10/14  10:16:59  hewett
 * Cleaned up header.
 *
**
** $Id: ctndisp.cc,v 1.12 2000/03/03 14:17:54 meichel Exp $
**		
*/

#include "osconfig.h" /* configure mechanism */
#include "ofstring.h" 
#include "cmdlnarg.h"
#include "ofconapp.h"
#include "dcuid.h"    /* for dcmtk version name */
#include "dcdebug.h"

#define OFFIS_CONSOLE_APPLICATION "ctndisp"

static char rcsid[] = "$dcmtk: " OFFIS_CONSOLE_APPLICATION " v"
  OFFIS_DCMTK_VERSION " " OFFIS_DCMTK_RELEASEDATE " $";

#ifdef HAVE_STDLIB_H
#ifndef  _BCB4
/* workaround for bug in Borland C++ Builder 4 */
BEGIN_EXTERN_C
#endif
#include <stdlib.h>
#ifndef  _BCB4
END_EXTERN_C
#endif
#endif

#include <stdio.h>

#ifndef X_DISPLAY_MISSING
#if defined(HAVE_SHMAT) && defined(HAVE_SHMCTL) && defined(HAVE_SHMGET)

#include <string.h>
#include <math.h>

BEGIN_EXTERN_C
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>

#ifdef HAVE_LIBC_H
#include <libc.h>
#endif

#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_SYS_IPC_H
#include <sys/ipc.h>
#endif

#ifdef HAVE_SYS_SHM_H
#include <sys/shm.h>
#else
/* 
   This is a dummy declaration of struct shmid_ds. The real declaration from 
   sys/shm.h is very complex. Since we only pass "opaque" pointers of this 
   type to shmctl() and don't need the results, this hack should work.
   Anyway: Which weard kind of system
   should support shmctl() and have no appropriate header file??? 
 */
     struct shmid_ds { int dummy_element; };
#endif

#ifndef HAVE_PROTOTYPE_SHMCTL
     int shmctl(int shmid, int cmd, struct shmid_ds *buf);
#endif

#ifdef HAVE_USLEEP
#ifndef HAVE_PROTOTYPE_USLEEP
     int usleep(unsigned int useconds);
#endif
#endif
END_EXTERN_C

#include "snq.h"
#include "disppriv.h"
#include "ctndsupp.h"
#include "ctnddico.h"
#include "dcdict.h"

/*
 * This typedef and these function prototypes occur here because gcc can't
 * handlel the X include libraries....this is the short answer...
 */
typedef struct _XImageStruct {
    XImage *x_image;		/* the X image struct		 */
    Window image_window;	/* the window for display	 */
}   XImageStruct;

void CTNDISP_DisplayImageData(ImageStruct * img, XImageStruct * ximg);
void CTNDISP_RefreshImageData();

/*
 * Globals (preceded with G_)...declared extern in other files
 */
Window G_big_window;
Display *G_display;		/* Xstuff used all over...	 */
Visual *G_visual;
Colormap G_cmap;

GC  G_gclargeT,
    G_gcsmallT,
    G_gcsmallTbold,
    G_gcblack,			/* Graphics contexts needed...	 */
    G_gcred,
    G_gcgreen;

int G_display_width = -1; /* -1's indicate to set the width and height to full screen */
int G_display_height = -1; /* -1's indicate to set the width and height to full screen */
int G_screen,
    G_ncolors;

int G_FirstTime = 1,
    Rescaleflag = 1,
    Windowflag = 1;		/* if to use the rescale and window data.
				 * default is yes */
char G_nodename[MAX_STRING_LENGTH];

QUEUE_Element Ctndisplayhandle,
    CalcElement;		/* CalcElement is the Image element in
				 * calculation (if MULTI) */

int MULTI = 1;
int ImageQueueStatus = EMPTY;
int CalcStatus = IFREE;

unsigned char *ishmaddr,
   *tishmaddr;			/* shared memory adress pointer */
int status,
    died,
    pid,
    shmflag;			/* status of son */
struct shmid_ds *shm_buf;
key_t ishmkey;
int ishmsize,
    ishmflag,
    ishmID;


/*
 * End Globals...
 */

ImageStruct cur_img,
    next_img;

XImageStruct xcur_img,
    xnext_img;

Image_Modul tempimage;


/*---------------------*/
void
InitSemaphore(int study_Q_id)
{

    unsigned long retval;

#ifdef DEBUG
    printf("Init Semaphore\n");
#endif
    /*
     * Initialize the SNQ for studies and status messages 
     */

    retval = SNQ_InitQueue(study_Q_id, QUEUE_SIZE);
#ifdef DEBUG
    printf("SNQ_InitQueue return value = %ld\n", retval);
#endif

    if (retval != SNQ_NORMAL) {
	fprintf(stderr,"CTNDISP: Not able to access Study-Name Queue.\n");
	exit(-1);
    }
} /* end InitSemaphore */


unsigned long
GetElement(QUEUE_Element * tempelement)
{
    return SNQ_Dequeue(tempelement);
}

#define SHORTCOL 4
#define LONGCOL 21

/*******************************************************************/
int main(int argc, char **argv)
{
	
  SetDebugLevel(0);
  strcpy(G_nodename, "CTNDISP");  
  
  int study_Q_id = 0;
  OFCmdUnsignedInt opt_windowWidth = 0;
  OFCmdUnsignedInt opt_windowHeight = 0;
  OFCmdUnsignedInt opt_queue = 0;
  const char *opt_nodename = NULL; 

  OFConsoleApplication app(OFFIS_CONSOLE_APPLICATION , "Visual monitor for imagectn", rcsid);
  OFCommandLine cmd;
  char opt_string[256];
  cmd.setOptionColumns(LONGCOL, SHORTCOL);
  cmd.setParamColumn(LONGCOL + SHORTCOL + 4);

  cmd.addParam("name",  "machine name to be displayed");
  
  cmd.addGroup("general options:", LONGCOL, SHORTCOL + 2);
   cmd.addOption("--help",                      "-h",        "print this help text and exit");
 
  cmd.addGroup("display options:");
    cmd.addSubGroup("number of images on display:", LONGCOL, SHORTCOL);
      cmd.addOption("--four-images",            "-i4",       "display four images (default)");
      cmd.addOption("--two-images",             "-i2",       "display two images");
    cmd.addSubGroup("window size:");
      cmd.addOption("--full-screen",            "-wf",       "use full screen (default)");
      cmd.addOption("--quarter-screen",         "-wq",       "use only quarter of screen");
      sprintf(opt_string, "[w]idth, [h]eight: integer (min. %d)", (int) MIN_DISPLAY_WIDTH);
      cmd.addOption("--size",                   "-ws",    2, opt_string, "set window size to w by h pixels");
    cmd.addSubGroup("image processing options:");
      cmd.addOption("--ignore-rescale",         "-ir",        "ignore rescale slope/intercept in images");
      cmd.addOption("--ignore-window",          "-iw",        "ignore window center/width in images");
  cmd.addGroup("other options:", LONGCOL, SHORTCOL + 2);
      cmd.addOption("--single-process",                       "do not use multiprocessing");
      cmd.addOption("--queue",                  "-q",      1, "[q]ueue-number: integer", "use queue number q (default: 0)");

    /* evaluate command line */                           
    prepareCmdLineArgs(argc, argv, OFFIS_CONSOLE_APPLICATION);
    if (app.parseCommandLine(cmd, argc, argv, OFCommandLine::ExpandWildcards))
    {
      cmd.getParam(1, opt_nodename);
      strncpy(G_nodename, opt_nodename, MAX_STRING_LENGTH-1);
      G_nodename[MAX_STRING_LENGTH-1]=0;
      
      cmd.beginOptionBlock();
      if (cmd.findOption("--four-images")) /* nothing */ ;
      if (cmd.findOption("--two-images"))
      {
        G_display_height = -3;	/* indicates in initXwindows that to use the half screen height */
	Set_Images_and_Show(2, 10);
      }
      cmd.endOptionBlock();

      cmd.beginOptionBlock();
      if (cmd.findOption("--full-screen")) /* nothing */ ;
      if (cmd.findOption("--quarter-screen"))
      {
        G_display_height = -2;
        G_display_width = -2;	/* indicates in initXwindows that to use the half screen height and width */
      }
      if (cmd.findOption("--size")) 
      {
        app.checkValue(cmd.getValue(opt_windowWidth, MIN_DISPLAY_WIDTH));
        app.checkValue(cmd.getValue(opt_windowHeight, MIN_DISPLAY_HEIGHT));
        G_display_width = (int)opt_windowWidth;
        G_display_height = (int)opt_windowHeight;
      }
      cmd.endOptionBlock();

      if (cmd.findOption("--ignore-rescale")) Rescaleflag=0 ;
      if (cmd.findOption("--ignore-window")) Windowflag=0 ;      
      if (cmd.findOption("--single-process")) MULTI=0 ;
      if (cmd.findOption("--queue")) 
      {
        app.checkValue(cmd.getValue(opt_queue, 0));
        study_Q_id = (int)opt_queue;
      }

    }

    unsigned long retval;
    QUEUE_Element element;
    XEvent event;
    memset(&tempimage, 0, sizeof(Image_Modul));

    /* make sure data dictionary is loaded */
    if (!dcmDataDict.isDictionaryLoaded()) {
	fprintf(stderr, "Warning: no data dictionary loaded, check environment variable: %s\n",
		DCM_DICT_ENVIRONMENT_VARIABLE);
    }

    /*
     * Initialize the SNQ for studies and status messages 
     */
    InitSemaphore(study_Q_id);

    /*
     * Initialize the X stuff, load gray-scale color table 
     */
    if (! CTNDISP_InitializeX(220)) {
	SNQ_KillQueue(study_Q_id);
        /* error msg already printed from CTNDISP_InitializeX */
	exit(-1);
    }
    /*
     * Display node name and the status buttons... 
     */
    CTNDISP_CreateShow();	/* do all the init stuff */


    /*
     * Main Control loop for ctndisp 
     *
     */

    /* NEW STUFF J.Burkert */
    int forever = 1; /* hack to prevent warnings on certain compilers */
    while (forever) {

	/* printf("%d events pending\n",XPending(G_display)); */

	if (XPending(G_display) > 0) {	/* first look if there is a xwindow
					 * event, if not examine the QUEUE
					 * else sleep a while */
	    XNextEvent(G_display, &event);

	    /* printf("event number : %d\n",event.type); */
	    switch (event.type) {
	    case ConfigureNotify:
		/* printf("Config notify\n"); */
		break;
	    case Expose:
		if (event.xexpose.count != 0)
		    break;	/* to avoid redraw often in a row */
		CTNDISP_RefreshImageData(event);

		CTNDISP_RefreshStatus();
		break;
	    case UnmapNotify:

	    case DestroyNotify:
		if (event.xdestroywindow.window == G_big_window) {
		    if ((SNQ_KillQueue(study_Q_id)) != SNQ_NORMAL) 
			fprintf(stderr, "CTNDISP: Not able to kill Study-Name Queue.\n");
		    if (MULTI) {
			if ((shmctl(ishmID, IPC_RMID, shm_buf)) == -1) {
			    shared_error();
			}
		    }
		    printf("CTNDISPLAY terminated. Bye \n");
		    exit(0);
		}
		break;
	    default:
		break;

	    } /* switch */
	}
	/* if Pending */
	else {
	    retval = GetElement(&element);
	    if (retval == SNQ_NORMAL) {


#ifdef DEBUG
		printf("show it \n");
#endif
		ShowElement(element);

		/* CTNDISP_RefreshStatus(); */
	    } else if (MULTI) {
		if (CalcStatus == ICALC) {
		    status = 0;
		    died = (int)waitpid(pid, &status, WNOHANG | WUNTRACED);
		    if ((status >> 8) == 42) {
			died = (int)wait(&status);

#ifdef DEBUG
			printf("IREADY in ctndisp.c \n");
			printf("%d WIFEXITED\n", WIFEXITED(status));
			printf("%d WEXITSTATUS\n", WEXITSTATUS(status));
			printf("%d WSTOPSIG\n", WSTOPSIG(status));
#endif

			CalcStatus = IREADY;
			ShowElement(CalcElement);
			CalcStatus = IFREE;
		    }
		} /* CALC */
		if ((CalcStatus == IFREE) && (ImageQueueStatus == NOTEMPTY)) {
#ifdef DEBUG
		    printf("IFREE in ctndisp.c \n");
#endif

		    CalcStatus = INEXT;
		    ImageQueue(&CalcElement, GET);	/* get image from queue */

		    ShowElement(CalcElement);
#ifdef DEBUG
		    printf("ICALC in ctndisp.c \n");
#endif

		    /*
		     * Calctstatus = icalc is set in ShowImage 
		     */
		} else {	/* default */
#ifdef HAVE_USLEEP
		    usleep(10000);	/* sleep 0.01 sec */
#else
		    sleep(1);
#endif
		} /* if there are no importent data queued */
	    } else {		/* if MULTI */
#ifdef HAVE_USLEEP
		usleep(10000);	/* sleep 0.01 sec */
#else
		sleep(1);
#endif
	    }

	} /* else */
    } /* while */
    return 0; /* this line is never reached but prevents a warning on some compilers. */
}

#else /* not HAVE_SHMxxx */

int main()
{
    fprintf(stderr, 
      "%s\n\n"
      "ctndisp requires shared memory (system calls: shmat, shmctl, shmget).\n"
      "configure detected that your system does not support shared memory.\n"
      "Please reconfigure your system and recompile if appropriate.\n",
      rcsid);
    return 0;
}

#endif
#else /* X_DISPLAY_MISSING */

int main()
{
    fprintf(stderr, 
      "%s\n\n"
      "ctndisp requires the X Windows System.\n"
      "configure detected that your system does not support X11.\n"
      "Please reconfigure your system and recompile if appropriate.\n",
      rcsid);
    return 0;
}

#endif /* X_DISPLAY_MISSING */
